"""
Message Module
"""

from email import message_from_bytes
from email.header import decode_header
from email.utils import parsedate_to_datetime

from email_profile.data import DataClass as Data
from email_profile.models import AttachmentModel, EmailModel


class Message:

    def __init__(self, message: bytes, id: int) -> None:
        self.message = message_from_bytes(message)
        self.id = id
        self.body_text_plain = ""
        self.body_text_html = ""
        self.data = Data()

    def decode_field(self, header) -> str:
        field = ""
        for sub in decode_header(header):
            encoding = sub[1] or "utf-8"
            try:
                field += sub[0].decode(encoding)
            except Exception:
                field += sub[0]
        return field

    def parsedate_to_datetime(self, date):
        try:
            return parsedate_to_datetime(date)
        except ValueError:
            pass
        return None

    def get_content(self, part) -> None:
        content_type = part.get_content_type()

        if content_type == "text/plain":
            try:
                self.body_text_plain = part.get_payload(decode=True).decode()
            except Exception:
                self.body_text_plain = part.get_payload(decode=True)

        if content_type == "text/html":
            try:
                self.body_text_html = part.get_payload(decode=True).decode()
            except Exception:
                self.body_text_html = part.get_payload(decode=True)
                if isinstance(self.body_text_html, bytes):
                    self.body_text_html = part.get_payload()

        if "attachment" in str(part.get("Content-Disposition")):
            filename = part.get_filename()
            if filename:
                self.data.add_attachment(
                    AttachmentModel(
                        email_id=self.id,
                        file_name=filename,
                        content_type=part.get_content_type(),
                        content_ascii=part.get_payload().encode("ascii")
                    )
                )

    def result(self) -> EmailModel:
        for part in self.message.walk():
            self.get_content(part=part)

        self.data.add_email(
            EmailModel(
                id=self.id,
                body_text_plain=self.body_text_plain,
                body_text_html=self.body_text_html,
                return_path=self.message.get("Return-Path"),
                delivered_to=self.message.get("Delivered-To"),
                received=self.message.get("Received"),
                dkim_signature=self.message.get("DKIM-Signature"),
                content_type=self.message.get_content_type(),
                date=self.parsedate_to_datetime(self.message.get("Date")),
                from_who=self.decode_field(self.message.get("From")),
                mime_version=self.message.get("Mime-Version"),
                message_id=self.message.get("Message-ID"),
                subject=self.decode_field(self.message["Subject"]),
                reply_to=self.message.get("Reply-To"),
                precedence=self.message.get("Precedence"),
                x_sg_eid=self.message.get("X-SG-EID"),
                x_sg_id=self.message.get("X-SG-ID"),
                to_who=self.message.get("To"),
                x_entity_id=self.message.get("X-Entity-ID")
            )
        )

        return self.data



"""
Utils Module
"""

from typing import List


class StatusResponse:

    def __init__(self, type: bool, message: str) -> None:
        self.type = type
        self.message = message


class Status:
    """
    ------
    URL: https://datatracker.ietf.org/doc/html/rfc3501#section-6.2.3
    ------

    Result:
        OK - login completed, now in authenticated state
        NO - login failure: user name or password rejected
        BAD - command unknown or arguments invalid
    """
    OK = 'OK'
    NO = 'NO'
    BAD = 'BAD'

    @staticmethod
    def validate_status(status: str, raise_error: bool = True) -> StatusResponse:
        data = {
            Status.OK: (True, "Login completed, now in authenticated state"),
            Status.NO: (False, "Login failure: user name or password rejected"),
            Status.BAD: (False, "Command unknown or arguments invalid")
        }.get(status, (False, "Unknown error"))

        if raise_error:
            if not data[0]:
                raise Exception(data[1])

        return StatusResponse(
            type=data[0],
            message=data[1]
        )

    @staticmethod
    def validate_data(data: List[str]) -> List[str]:
        data = [x.decode() for x in data[0].split()]
        if len(data) == 1 and not data[0]:
            return []
        return data


class Mailbox:

    INBOX = "INBOX"
    SENT = "INBOX.Sent"
    JUNK = "INBOX.Junk"
    DRAFTS = "INBOX.Drafts"


class Mode:

    ALL = "ALL"
    UNSEEN = "UNSEEN"



from email_profile.data import DataClass
from email_profile.models import AttachmentModel, EmailModel


def test_instance_data_sqlalchemy():
    data = DataClass()

    assert isinstance(data, DataClass)
    assert data.email is None
    assert data.attachments == list()


def test_data_sqlalchemy_add_email():
    email = EmailModel()
    data = DataClass()
    data.add_email(model=email)

    assert isinstance(data.email, EmailModel)
    assert data.email == email


def test_data_sqlalchemy_add_attachment():
    attachment = AttachmentModel()
    data = DataClass()
    data.add_attachment(model=attachment)

    assert len(data.attachments) == 1
    assert isinstance(data.attachments[0], AttachmentModel)
    assert data.attachments[0] == attachment


def test_data_sqlalchemy_json():
    email = EmailModel(id=1)
    attachment = AttachmentModel(id=42)

    data = DataClass()
    data.add_email(model=email)
    data.add_attachment(model=attachment)

    response = data.json()

    assert isinstance(response, dict)
    assert response.get("email")
    assert response.get("email")["id"] == 1
    assert response.get("attachments")[0]["id"] == 42



from email_profile.models.attachment import AttachmentModel
from email_profile.models.email import EmailModel
from email_profile.models.mailbox import MailBoxModel

__all__ = [
    "AttachmentModel",
    "EmailModel",
    "MailBoxModel",
]



from email_profile.utils import Status, StatusResponse


def test_status_ok():
    assert Status.OK == 'OK'


def test_status_no():
    assert Status.NO == 'NO'


def test_status_bad():
    assert Status.BAD == 'BAD'


def test_status_status_response():
    assert_type = True
    assert_message = "Ok!"

    status_response = StatusResponse(
        type=assert_type,
        message=assert_message
    )

    assert status_response.type is assert_type
    assert status_response.message == assert_message


def test_status_instance_validate_status():
    input_status = 'OK'
    validate = Status.validate_status(input_status)

    assert isinstance(validate, StatusResponse)


def test_status_validate_status_ok():
    input_status = 'OK'
    validate = Status.validate_status(input_status)

    assert validate.type is True
    assert validate.message == "Login completed, now in authenticated state"


def test_status_validate_status_no():
    input_status = 'NO'
    validate = Status.validate_status(
        status=input_status,
        raise_error=False
    )

    assert validate.type is False
    assert validate.message == "Login failure: user name or password rejected"


def test_status_validate_status_bad():
    input_status = 'BAD'
    validate = Status.validate_status(
        status=input_status,
        raise_error=False
    )

    assert validate.type is False
    assert validate.message == "Command unknown or arguments invalid"


def test_status_validate_data_with_data():
    input_status = [b'4 8 15 16 23 42']
    output_status = ['4', '8', '15', '16', '23', '42']

    assert Status.validate_data(input_status) == output_status


def test_status_validate_data_no_data():
    input_status = [b'']
    output_status = []

    assert Status.validate_data(input_status) == output_status


def test_status_validate_status_with_error_no():
    input_status = 'NO'
    try:
        Status.validate_status(
            status=input_status,
            raise_error=True
        )
    except Exception as error:
        current_error = error.args[0]

    assert current_error == "Login failure: user name or password rejected"


def test_status_validate_status_with_error_bad():
    input_status = 'BAD'
    try:
        Status.validate_status(
            status=input_status,
            raise_error=True
        )
    except Exception as error:
        current_error = error.args[0]

    assert current_error == "Command unknown or arguments invalid"


def test_status_validate_status_with_error_unknown():
    input_status = ''
    try:
        Status.validate_status(
            status=input_status,
            raise_error=True
        )
    except Exception as error:
        current_error = error.args[0]

    assert current_error == "Unknown error"


def test_status_validate_status_no_error():
    input_status = ''

    validate = Status.validate_status(
        status=input_status,
        raise_error=False
    )

    assert validate.type is False
    assert validate.message == "Unknown error"



"""
Where Module
"""


import logging

from datetime import date
from typing import Optional, List

from email_profile.config.controller import (
    AttachmentController,
    EmailController
)
from email_profile.serializers import WhereSerializer
from email_profile.utils import Status, Mode, Mailbox
from email_profile.message import Message


class Where:

    _data = []
    _message = []
    _status = False
    _total = 0

    def __init__(self,
                 mode: Mode = Mode.ALL,
                 mailbox: Mailbox = Mailbox.INBOX,
                 server: any = None) -> None:
        self.mode = mode
        self.mailbox = mailbox
        self.server = server

    def where(self,
              since: Optional[date] = None,
              before: Optional[date] = None,
              subject: Optional[str] = None,
              from_who: Optional[str] = None) -> object:
        variables = locals().copy()
        options = {}

        for item in variables:
            if variables[item] and item != 'self':
                options[item] = variables[item]

        status, total = self.server.select(self.mailbox.capitalize())
        validate = Status.validate_status(status)
        self._status = validate.type
        self._total = int(total[0].decode())

        status, data = self.server.search(None, WhereSerializer(**options).result())
        validate = Status.validate_status(status)
        self._status = validate.type
        self._data = Status.validate_data(data)

        return self

    def count(self) -> int:
        return len(self._data)

    def list_id(self) -> List[str]:
        return self._data

    def list_data(self) -> List[any]:
        if self._data:
            _sum = 1
            _sum_searching = 0
            _groups = 1

            while _sum < len(self._data):
                _sum += 100
                _groups += 1

            splited = [self._data[item::_groups] for item in range(_groups)]

            for group_mail in splited:
                _sum_searching += len(group_mail)

                status, messages = self.server.fetch(','.join(group_mail), '(RFC822)')
                messages = [message for message in messages if message != b')']

                print(f"Searching: {_sum_searching}/{len(self._data)}")

                for reference, text in messages:
                    _id = int(reference.split()[0])
                    self._message.append(Message(text, _id).result())

        return self._message

    def dump_sqlite(self):
        logging.warning(" Function 'dump_sqlite' not implemented")

        sql_email = EmailController()
        sql_attachmentl = AttachmentController()

        for message in self._message:
            sql_email.create(data=message.email)

            for attachment in message.attachments:
                sql_attachmentl.create(data=attachment)



from datetime import datetime, date
from email_profile.serializers import WhereSerializer


def test_instance_where_serializer():
    data = WhereSerializer()

    assert isinstance(data, WhereSerializer)
    assert data.since == ''
    assert data.before == ''
    assert data.subject == ''
    assert data.from_who == ''


def test_where_serializer_since():
    test_date = datetime(1996, 5, 31)
    assert_date = test_date.strftime('%d-%b-%Y')

    serializer = WhereSerializer(since=test_date)
    assert serializer.since == f'(SINCE {assert_date})'


def test_where_serializer_since_wrong_data_type():
    test_date = "1996-5-31"
    test_assert = "Attribute validation error (SINCE)."

    try:
        WhereSerializer(since=test_date)
    except AttributeError as error:
        current_error = error.args[0]

    assert current_error == test_assert


def test_where_serializer_before():
    test_date = date.today()
    assert_date = test_date.strftime('%d-%b-%Y')

    serializer = WhereSerializer(before=test_date)
    assert serializer.before == f'(BEFORE {assert_date})'


def test_where_serializer_before_wrong_data_type():
    test_date = "1996-5-31"
    test_assert = "Attribute validation error (BEFORE)."

    try:
        WhereSerializer(before=test_date)
    except AttributeError as error:
        current_error = error.args[0]

    assert current_error == test_assert


def test_where_serializer_subject_no_special_characters():
    test_string = "Important email"
    assert_string = "Important email"

    serializer = WhereSerializer(subject=test_string)
    assert serializer.subject == f'(SUBJECT "{assert_string}")'


def test_where_serializer_subject_with_special_characters():
    test_string = "Important êmail"
    assert_string = "Important mail"

    serializer = WhereSerializer(subject=test_string)
    assert serializer.subject == f'(SUBJECT "{assert_string}")'


def test_where_serializer_from_who_no_special_characters():
    test_string = "email@test.com"
    assert_string = "email@test.com"

    serializer = WhereSerializer(from_who=test_string)
    assert serializer.from_who == f'(FROM "{assert_string}")'


def test_where_serializer_from_who_with_special_characters():
    test_string = "êmail@test.com"
    assert_string = "mail@test.com"

    serializer = WhereSerializer(from_who=test_string)
    assert serializer.from_who == f'(FROM "{assert_string}")'


def test_where_serializer_result():
    input_since = datetime(1996, 5, 31)
    input_before = date.today()
    input_subject = 'abc'

    output_since = input_since.strftime('%d-%b-%Y')
    output_before = input_before.strftime('%d-%b-%Y')
    output_subject = input_subject
    output_result = f'(SINCE {output_since}) (BEFORE {output_before}) (SUBJECT "{output_subject}")'

    serializer = WhereSerializer(
        since=input_since,
        before=input_before,
        subject=input_subject
    )

    assert serializer.result() == output_result



"""
Controller Module
"""


import logging

from email_profile.config.database import session
from email_profile.models import AttachmentModel, EmailModel


class Controller:

    def __init__(self, model=None) -> None:
        self.model = model
        self.fields = model.__dataclass_fields__.items()
        self.table_name = model.__tablename__

        self._validate_table()

    def _create_table(self) -> None:
        fields = [name for name, field in self.fields]
        query = ", ".join(fields)

        session.execute(f"CREATE TABLE {self.table_name} ({query})")

    def _validate_table(self) -> None:
        try:
            session.execute(f"SELECT * FROM {self.table_name}")
        except Exception as error:
            error_message = f'no such table: {self.table_name}'

            if error.args[0] == error_message:
                self._create_table()

    def create(self, data: object):
        logging.info("Not implemented")
        pass

    def read(self):
        logging.info("Not implemented")
        pass

    def update(self):
        logging.info("Not implemented")
        pass

    def delete(self):
        logging.info("Not implemented")
        pass


class AttachmentController(Controller):

    def __init__(self):
        super().__init__(model=AttachmentModel)


class EmailController(Controller):

    def __init__(self):
        super().__init__(model=EmailModel)



from dataclasses import dataclass, field


@dataclass
class AttachmentModel:

    __tablename__ = "attachment"

    id: int = field(default=None)
    email_id: int = field(default=None)
    file_name: str = field(default=None)
    content_type: str = field(default=None)
    content_ascii: str = field(default=None)



"""
Dataclass Module
"""

from datetime import date
from dataclasses import dataclass, field


class Validations:

    def __post_init__(self):
        for name, _field in self.__dataclass_fields__.items():
            if method := getattr(self, f"validate_{name}"):
                setattr(self, name, method(field=_field))


@dataclass
class WhereSerializer(Validations):

    since: date = field(default="")
    before: date = field(default="")
    subject: str = field(default="")
    from_who: str = field(default="")

    def validate_since(self, field) -> str:
        if self.since:
            if not isinstance(self.since, field.type):
                raise AttributeError(
                    "Attribute validation error (SINCE)."
                )
            return '(SINCE {})'.format(
                self.since.strftime('%d-%b-%Y')
            )
        return ''

    def validate_before(self, field) -> str:
        if self.before:
            if not isinstance(self.before, field.type):
                raise AttributeError(
                    "Attribute validation error (BEFORE)."
                )
            return '(BEFORE {})'.format(
                self.before.strftime('%d-%b-%Y')
            )
        return ''

    def validate_subject(self, field) -> str:
        if self.subject:
            if not isinstance(self.subject, field.type):
                raise AttributeError(
                    "Attribute validation error (SUBJECT)."
                )
            return '(SUBJECT "{}")'.format(
                self.subject.encode("ASCII", 'ignore').decode()
            )
        return ''

    def validate_from_who(self, field) -> str:
        if self.from_who:
            if not isinstance(self.from_who, field.type):
                raise AttributeError(
                    "Attribute validation error (FRON_WHO)."
                )
            return '(FROM "{}")'.format(
                self.from_who.encode("ASCII", 'ignore').decode()
            )
        return ''

    def result(self):
        data = []
        for _field in self.__dataclass_fields__:
            content = getattr(self, _field)
            if content:
                data.append(content)
        return " ".join(data)



"""
Data Module
"""

from typing import List, Dict
from abc import abstractmethod, ABC


class DataAbstract(ABC):

    def __init__(self) -> None:
        self.email: object = None
        self.attachments: List[object] = list()

    def add_email(self, model: object):
        self.email = model

    def add_attachment(self, model: object):
        self.attachments.append(model)

    @abstractmethod
    def json(self) -> Dict:
        return {
            "email": {},
            "attachments": [{}]
        }


class DataClass(DataAbstract):

    def json(self) -> Dict:
        attachments_temp = []
        for attachment in self.attachments:
            attachments_temp.append(attachment.__dict__)

        return {
            "email": self.email.__dict__,
            "attachments": attachments_temp
        }



"""
Core Module
"""

import imaplib
from functools import partial

from email_profile.where import Where


class Email:

    def __init__(self, server: str, user: str, password: str) -> None:
        self.server = imaplib.IMAP4_SSL(server)
        self.server.login(user=user, password=password)
        self.select = partial(Where, server=self.server)



from typing import Optional
from datetime import datetime
from dataclasses import dataclass, field


@dataclass
class EmailModel:

    __tablename__ = 'email'

    id: Optional[int] = field(default=None)
    body_text_plain: str = field(default=None)
    body_text_html: str = field(default=None)
    return_path: str = field(default=None)
    delivered_to: str = field(default=None)
    received: str = field(default=None)
    dkim_signature: str = field(default=None)
    received: str = field(default=None)
    content_type: str = field(default=None)
    date: datetime = field(default=None)
    from_who: str = field(default=None)
    mime_version: str = field(default=None)
    message_id: str = field(default=None)
    subject: str = field(default=None)
    reply_to: str = field(default=None)
    precedence: str = field(default=None)
    x_sg_eid: str = field(default=None)
    x_sg_id: str = field(default=None)
    to_who: str = field(default=None)
    x_entity_id: str = field(default=None)



from email_profile.utils import Mailbox


def test_mainbox_inbox():
    assert Mailbox.INBOX == 'INBOX'


def test_mainbox_sent():
    assert Mailbox.SENT == 'INBOX.Sent'


def test_mainbox_junk():
    assert Mailbox.JUNK == 'INBOX.Junk'


def test_mainbox_drafts():
    assert Mailbox.DRAFTS == 'INBOX.Drafts'



import os

from datetime import date, datetime
from email_profile import Email

try:
    from dotenv import load_dotenv
except Exception:
    pass


def main():
    """Exemple"""
    load_dotenv()

    app = Email(
        server=os.getenv("EMAIL_SERVER"),
        user=os.getenv("EMAIL_USERNAME"),
        password=os.getenv("EMAIL_PASSWORD")
    )

    # Query instance
    query = app.select(mailbox="Inbox").where(
        since=datetime(1996, 5, 31),
        before=date.today(),
        subject='abc'
    )

    # Count
    print(query.count())

    # List IDs
    ids = query.list_id()
    print(ids)

    # List Data
    data = query.list_data()

    for content in data:
        # Email data model
        print(content.email.subject)

        # Attachments data model
        print(content.attachments)

        # Dump Json
        print(content.json())


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from email_profile import __version__
from setuptools import setup
from setuptools.command.install import install


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


class CustomInstallCommand(install):
    def run(self):
        install.run(self)


setup(
    name="email-profile",
    fullname='email-profile',
    version=__version__,
    author="Fernando Celmer",
    author_email="email@fernandocelmer.com",
    description="📩 Email Profile",
    long_description=long_description,
    long_description_content_type="text/markdown",
    project_urls = {
        'Homepage': 'https://github.com/linux-profile/email-profile',
        'Repository': 'https://github.com/linux-profile/email-profile',
        'Documentation': 'https://github.com/linux-profile/email-profile/blob/master/README.md',
        'Issues': 'https://github.com/linux-profile/email-profile/issues',
    },
    cmdclass={
        'install': CustomInstallCommand,
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        'Intended Audience :: Developers',
        'Natural Language :: English',
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    packages=['email_profile'],
    include_package_data=True,
    python_requires=">=3.8",
    zip_safe=True,
    entry_points={
        'console_scripts': ['email=email_profile.main:main'],
    },
)
